// ==============================================================
//
//  Copyright (c) 2002-2003 Alex Vinokur.
//
//  ------------------------------------------------------------
//  This file is part of C++ Simulator of a Post Machine.
//
//  C++ Simulator of a Post Machine is free software; 
//  you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License,
//  or (at your option) any later version.
//
//  C++ Simulator of a Post Machine is distributed in the hope 
//  that it will be useful, but WITHOUT ANY WARRANTY; 
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with C++ Simulator of a Post Machine; 
//  if not, write to the Free Software Foundation, Inc., 
//  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ------------------------------------------------------------
// 
//  mailto:alexvn@connect.to
//  http://up.to/alexv
//
// ==============================================================




// ##############################################################
//
//  SOFTWARE : Nondeterministic Post Machine (C++ Simulator)
//  FILE     : commands.cpp
//
//  DESCRIPTION :
//         Class CommandAction (Implementation)
//
// ##############################################################




// ###############
#include "commands.h"


// =========
const name64_t cmd_type_name[] =
{
  "UNDEF_Command_Type",		// UNDEF_CommandType
  "SEQUENTIAL",			// SEQUENTIAL_cmd
  "COND-JUMP", 			// CONDITIONAL_JUMP_cmd
  "END" 			// END_cmd
};


const name64_t op_type_name[] =
{
  "UNDEF_OperationType",	// UNDEF_OperationType
  "STAY",			// STAY_ON_op
  "LEFT",			// LEFT_SHIFT_op
  "RIGHT",			// RIGHT_SHIFT_op
  "WRITE",			// WRITE_LABEL_op
  "ERASE",			// ERASE_LABEL_op
  "MARKED",			// CHECK_IF_MARKED_op
  "BLANK",			// CHECK_IF_BLANK_op
  "STOP"			// STOP_op
};


// =========
vector<string> CommandAction::op_names_s (op_type_name, op_type_name + sizeof(op_type_name)/sizeof(*op_type_name));


// =========
// =========
// Constructor-0
CommandAction::CommandAction ()
{
} 

// =========
// Constructor-1
CommandAction::CommandAction (
	const vector<OperationType>&  cur_operations_i,
	const vector<command_t>&      next_commands_i
	)
	:
	cur_operations_ (cur_operations_i),
	next_commands_ (next_commands_i)
{
} 


// =========
// Destructor
CommandAction::~CommandAction ()
{
} 


// =========
CommandType CommandAction::get_command_type_S (OperationType op_i)
{

  switch (op_i)
  {
    case STAY_ON_op :
    case LEFT_SHIFT_op :
    case RIGHT_SHIFT_op :
    case WRITE_LABEL_op :
    case ERASE_LABEL_op :
      return SEQUENTIAL_cmd;
      break;

    case CHECK_IF_MARKED_op :
    case CHECK_IF_BLANK_op  :
      return CONDITIONAL_JUMP_cmd;
      break;

    case STOP_op :
      return END_cmd;
      break;

    default :
      assert (0);
      return UNDEF_CommandType;   // unused
      break; // unused
  }

  assert (0);                     // unused
  return UNDEF_CommandType;       // unused
} 


// =========
bool CommandAction::is_allowed_op_name_S (const string& op_name_i)
{
  if (find (op_names_s.begin(), op_names_s.end(), op_name_i) == op_names_s.end()) return false;
  if (op_name_i == op_names_s[UNDEF_OperationType]) return false;
  return true;
}

// =========
string CommandAction::get_allowed_op_names_S (CommandType cmd_type)
{
ostringstream oss;
  for (size_t k = 0; k < op_names_s.size(); k++)
  {
    if (k == size_t (UNDEF_OperationType)) continue;

    const string op_name = op_names_s[k]; 

    if ((cmd_type == UNDEF_CommandType) || (get_command_type_S (get_op_type_S (op_name)) == cmd_type)) 
    {
      oss << op_name << " ";
    }
  }
  assert (!oss.str().empty());
  return oss.str().substr (0, oss.str().size() - 1);

}

size_t CommandAction::get_op_names_max_size_S ()
{
size_t ret_value = 0;
  for (size_t k = 0; k < op_names_s.size(); k++)
  {
    if (k == UNDEF_CommandType) continue;
    ret_value = MAX_VALUE (ret_value, op_names_s[k].size());
  }
  return ret_value;
}


// =========
OperationType CommandAction::get_op_type_S (const string& op_name_i)
{
  assert (is_allowed_op_name_S (op_name_i));

vector<string>::iterator find_iter = find (op_names_s.begin(), op_names_s.end(), op_name_i);
  assert (find_iter != op_names_s.end());
  assert (find_iter != op_names_s.begin()); // UNDEF_OperationType

  return (OperationType(distance (op_names_s.begin(), find_iter)));
}




// =========
string CommandAction::get_cmd_type_name_S (CommandType cmd_type)
{
const vector<string> cmd_names(cmd_type_name, cmd_type_name + sizeof(cmd_type_name)/sizeof(*cmd_type_name));

ostringstream oss;
  for (size_t k = 0; k < cmd_names.size(); k++)
  {
    if (k == size_t (UNDEF_CommandType)) continue;

    const string cmd_name = cmd_names[k]; 

    if ((cmd_type == UNDEF_CommandType) || (k == size_t (cmd_type))) 
    {
      oss << cmd_name << " ";
    }
  }
  assert (!oss.str().empty());
  return oss.str();

}


// =========
vector<command_t> CommandAction::get_next_commands () const
{
  return next_commands_;
}



// =========
CommandType CommandAction::get_command_type () const
{
  assert (!cur_operations_.empty());
  for (size_t k = 0; k < cur_operations_.size(); k++)
  {
    assert (get_command_type_S (cur_operations_[k]) == get_command_type_S (cur_operations_.front()));
  }
  return get_command_type_S (cur_operations_.front());
}


// =========
string CommandAction::getstr_cur_operation_names (size_t size_i) const
{
ostringstream oss;

  for (size_t k = 0; k < cur_operations_.size(); k++)
  {
    if (size_i > 0) oss << setw (size_i) << left;
    oss << op_names_s[cur_operations_[k]].c_str() << " ";
  }

  return oss.str();
}


// =========
string CommandAction::getstr_next_commands (size_t size_i) const
{
ostringstream oss;
  for (size_t k = 0; k < next_commands_.size(); k++)
  {
    if (size_i > 0) oss << setw (size_i) << right;
    oss << next_commands_[k] << " ";
  }
  return oss.str();

}


