// ==============================================================
//
//  Copyright (c) 2002-2003 Alex Vinokur.
//
//  ------------------------------------------------------------
//  This file is part of C++ Simulator of a Post Machine.
//
//  C++ Simulator of a Post Machine is free software; 
//  you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License,
//  or (at your option) any later version.
//
//  C++ Simulator of a Post Machine is distributed in the hope 
//  that it will be useful, but WITHOUT ANY WARRANTY; 
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with C++ Simulator of a Post Machine; 
//  if not, write to the Free Software Foundation, Inc., 
//  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ------------------------------------------------------------
// 
//  mailto:alexvn@connect.to
//  http://up.to/alexv
//
// ==============================================================




// ##############################################################
//
//  SOFTWARE : Nondeterministic Post Machine (C++ Simulator)
//  FILE     : post-m.cpp
//
//  DESCRIPTION :
//         Class NondeterministicPostMachine (Implementation)
//
// ##############################################################


// ###############
#include "post-m.h"


#define	SETW_RULE_NUMBER	4
#define	UNBELIEVABLE_NUMBER_OF_TAPES	100


// =========
size_t	NondeterministicPostMachine::counter_s (0);
size_t	MachineStatus::next_id_to_be_processed_s (0);


// =========
// =========
// Constructor-0
MachineStatus::MachineStatus ()	: the_configuration_no_(0), the_command_subno_ (0)
{
}
 

// =========
// Destructor
MachineStatus::~MachineStatus ()
{
} 
		      

// =========
// =========
// Constructor-0
NondeterministicPostMachine::NondeterministicPostMachine (
	size_t	                       number_of_tapes_i,
	const vector<vector<string> >& descr_i, 
	const vector<command_t>&       initial_command_numbers_i,
	const vector<command_t>&       halting_command_numbers_i,
	const vector<command_t>&       internal_command_numbers_i,
	const Commands_t&              commands_i,
	const string&                  msg_i
	)
	:
	is_actually_nondeterministic_ (false),
	descr_ (descr_i),
	initial_commands_ (initial_command_numbers_i),
	halting_commands_ (halting_command_numbers_i),
	internal_commands_ (internal_command_numbers_i),
	commands_ (commands_i),
	check_results_ (true),
	cur_input_set_no_ (0),
        command_statistics_ (commands_.size()),
	max_command_size_ (0),
	max_symbol_size_ (0),
	serial_no_ (++counter_s),
	input_size_ (0),
	output_size_ (0),
	space_used_ (0),
	time_used_ (0)

{

  IF_NOT_EMPTY (msg_i, 5, '=');

typedef Tapes_t::value_type value_type;
pair<Tapes_t::iterator, bool> tapes_pair;

  if (number_of_tapes_i == 0)
  {
    check_results_ = false;
    FATAL_MSG	("Illegal number of tapes : "
		<< number_of_tapes_i
		);
    return;
  }

  if (number_of_tapes_i > UNBELIEVABLE_NUMBER_OF_TAPES)
  {
    WARNING_MSG	("Unbelievable number of tapes : "
		<< number_of_tapes_i
		);
  }

  // -----------------------------------
  for (size_t i = 0; i < number_of_tapes_i; i++)
  {
    tapes_pair = init_tapes_.insert (value_type(i, Tape ()));
    if (!tapes_pair.second)
    {
      check_results_ = false;
      FATAL_MSG	("The same current position occurs more than once in command rules");
      break;
    }
    assert (tapes_pair.second);

  }

  // ------------------------------------
  set_max_command_size();
  set_max_symbol_size();


  // ----------------------
  if (check_results_)	check_results_ = check_commands ();
  if (check_results_)	check_results_ = check_command ();

  assert (!commands_.empty());
  detect_if_is_actually_nondterministic ();

  // ---------------
  show_env();

} 

// =========
// Destructor
NondeterministicPostMachine::~NondeterministicPostMachine ()
{
} 

// =========
void NondeterministicPostMachine::set_max_command_size()
{
vector<command_t>	tmp_all_commands = get_all_commands();
  assert (max_command_size_ == 0);
  for (size_t i = 0; i < tmp_all_commands.size(); i++)
  {
    max_command_size_ = MAX_VALUE (max_command_size_, to_string(tmp_all_commands[i]).size()); 
  }

}

// =========
void NondeterministicPostMachine::set_max_symbol_size()
{
Tapes_t::const_iterator pos_iter;
  assert (max_symbol_size_ == 0);
  for (pos_iter = init_tapes_.begin(); pos_iter != init_tapes_.end(); pos_iter++)
  {
    max_symbol_size_ = MAX_VALUE (max_symbol_size_, pos_iter->second.max_symbol_size_);
  }
}

// =========
vector<command_t> NondeterministicPostMachine::get_all_commands () const
{
vector<command_t>	ret_vector;

  copy (initial_commands_.begin(), initial_commands_.end(), back_inserter(ret_vector));
  copy (halting_commands_.begin(), halting_commands_.end(), back_inserter(ret_vector));
  copy (internal_commands_.begin(), internal_commands_.end(), back_inserter(ret_vector));

  return ret_vector;

}


// =========
void NondeterministicPostMachine::set_input (
		const vector<vector<symbol_t> >& input_words_i,
		const vector<long>&              init_pos_i
		)

{
  assert (!cur_status_.empty());
  get_cur_command_no() = initial_commands_.front();

Tapes_t::iterator pos_iter;
  for (pos_iter = get_cur_tapes().begin(); pos_iter != get_cur_tapes().end(); pos_iter++)
  {
    pos_iter->second.clear_it();
  }
  get_cur_tapes().clear();

  get_cur_tapes() = init_tapes_;


  for (size_t i = 0; i < get_cur_tapes().size(); i++) 
  {
    get_cur_tapes()[i].clear_it();
  }

  for (size_t i = 0; i < input_words_i.size(); i++) 
  {
    get_cur_tapes()[i].set_input(input_words_i[i], init_pos_i[i]);
  }

  // --------------------------
  assert (input_size_ == 0);
  for (size_t i = 0; i < input_words_i.size(); i++) 
  {
    if (
         (input_words_i[i].size() == 1) 
         && 
         (get_cur_tapes()[i].is_empty_symbol(input_words_i[i].back()))
       ) 
    {
     continue;
    }

    input_size_ += input_words_i[i].size();
  }

}



// =========
ulong NondeterministicPostMachine::get_output_size ()
{
ulong ret_value = 0;

  for (size_t i = 0; i < get_cur_tapes().size(); i++) 
  {
    ret_value += get_cur_tapes()[i].get_output_word ().size();
  }

  return ret_value;
}



// =========
ulong NondeterministicPostMachine::get_space_used ()
{
ulong ret_value = 0;

  for (size_t i = 0; i < get_cur_tapes().size(); i++) 
  {
    ret_value += get_cur_tapes()[i].cell_visits_.size();
  }

  return ret_value;
}


// =========
void NondeterministicPostMachine::clear_it ()
{
  cur_status_.clear();
  MachineStatus::next_id_to_be_processed_s = 0;
  cur_status_.push_back (MachineStatus());
  assert (!cur_status_.empty());

  cur_input_set_no_++;

  input_size_   = 0;
  output_size_  = 0;
  space_used_   = 0;
  time_used_    = 0;

  for (size_t i = 0; i < command_statistics_.size(); i++)
  {
    command_statistics_[i] = 0;
  } 

}


// =========
bool	NondeterministicPostMachine::check_input_words (
		vector<vector<symbol_t> >& input_words_io,
		vector<long>& init_pos_o
		) const

{
Tapes_t::const_iterator	pos_iter;

  if (input_words_io.empty())
  {
    FATAL_MSG	("No input words");
    return false;
  }

  init_pos_o.clear();
  init_pos_o.resize(input_words_io.size(), HEAD_START_POSITION_DEFAULT);
  assert (init_pos_o.size() == input_words_io.size());

  if (input_words_io.size() != init_tapes_.size())
  {
    FATAL_MSG	("Number of input words ("
		<< input_words_io.size()
		<< ") is NOT equal number of tapes ("
		<< init_tapes_.size()
		<< ")"
		);
    return false;  
  }

size_t i;
  for (i = 0, pos_iter = init_tapes_.begin(); i < input_words_io.size(); i++, pos_iter++)
  {
    assert (pos_iter->first == i);
    if (input_words_io[i].empty())
    {
      FATAL_MSG	("No input word for tape#"
		<< i
		);
      return false;
    }


    vector<symbol_t>::iterator find_iter;
    // ---------------------------------
    const vector<symbol_t> original_input_word (input_words_io[i]);
    find_iter = find (input_words_io[i].begin(), input_words_io[i].end(), HEAD_START_POSITION_POINTER_REDERVED_SYMBOL); 
    if (find_iter != input_words_io[i].end())
    {
      init_pos_o [i] = long (distance (input_words_io[i].begin(), find_iter));
      input_words_io[i].erase(find_iter);
    }

    find_iter = find (input_words_io[i].begin(), input_words_io[i].end(), HEAD_START_POSITION_POINTER_REDERVED_SYMBOL); 
    if (find_iter != input_words_io[i].end())
    {
      ostringstream oss;
      copy (original_input_word.begin(), original_input_word.end(), ostream_iterator<symbol_t> (oss, " "));
      FATAL_MSG	("Tape#"
		<< i
		<< ", input word <"
		<< oss.str()
		<< ">" 
		<< " contains more than one reserved symbol "
		<< "\""
		<< HEAD_START_POSITION_POINTER_REDERVED_SYMBOL
		<< "\""
		<< "; Note! That symbol is used to point at initial position of the machine head"
		);

      return false;
    }



    // ####################################################
    for (size_t j = 0; j < input_words_io[i].size(); j++)
    {
      const bool is_label = pos_iter->second.is_label_symbol (input_words_io[i][j]);
      const bool is_blank = pos_iter->second.is_empty_symbol (input_words_io[i][j]);
      if (!(is_label || is_blank))
      {
         FATAL_MSG ("Tape#"
                   << i
                   << ", symbol#"
                   << j
                   << " : "
                   << "\""
                   << input_words_io[i][j]
                   << "\""
                   << " is NOT an input or blank symbol : "
                   << LABEL_ALPHABET_SYMBOL
                   << " or "
                   << EMPTY_ALPHABET_SYMBOL
                   );
          return false;
      }
    } // for (size_t j = 0; ...
    // ####################################################

  } // for (size_t i = 0; ...

  return true;
}


// =========
void NondeterministicPostMachine::show_input (
	const vector<vector<symbol_t> >& input_words_i,
	const vector<long>&              init_pos_i,
	bool                             fool_flag_i
	) const
{
  cout << endl;
  cout << "\t"
       << string (5, '#')
       << " ";
  if (fool_flag_i)
  {
    cout << ""
         << getstr_id_info ();
  }
  cout << "Input word(s) & head start position(s) on tape(s)"
       << " " 
       << string (5, '#')
       << endl;

  assert (input_words_i.size() == init_pos_i.size());

const size_t number_of_tapes = input_words_i.size();
  for (size_t i = 0; i < number_of_tapes; i++)
  {
    cout << "Tape#" << i << " : Word = ";
    for (size_t j = 0; j < input_words_i[i].size(); j++)
    {
      cout << setw (max_symbol_size_) << to_string (input_words_i[i][j]).c_str() << " ";
    }
    cout << ";  Head pos = " << init_pos_i [i];
    //copy (input_words_i[i].begin(), input_words_i[i].end(), ostream_iterator<symbol_t> (cout, " "));
    cout << endl;
  }
  cout << endl;
}



// =========
size_t& NondeterministicPostMachine::get_cur_configuration()
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_configuration_no_;
}



// =========
size_t NondeterministicPostMachine::get_processed_configuration() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_configuration_no_;
}



// =========
Tapes_t& NondeterministicPostMachine::get_cur_tapes()
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_tapes_;
}


// =========
Tapes_t NondeterministicPostMachine::get_processed_tapes() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_tapes_;
}


// =========
command_t& NondeterministicPostMachine::get_cur_command_no()
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_command_no_;
}



// =========
command_t NondeterministicPostMachine::get_processed_command_no() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_command_no_;
}



// =========
Commands_t::const_iterator& NondeterministicPostMachine::get_cur_command_iter()
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_command_iter_;
}



// =========
Commands_t::const_iterator NondeterministicPostMachine::get_processed_command_iter() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_command_iter_;
}



// =========
int& NondeterministicPostMachine::get_cur_command_subno()
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_command_subno_;
}



// =========
int NondeterministicPostMachine::get_processed_command_subno() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_command_subno_;
}



// =========
vector<int> NondeterministicPostMachine::get_cur_branch() const
{
  assert (!cur_status_.empty());
  return cur_status_.back().the_branch_;
}



// =========
vector<int> NondeterministicPostMachine::get_processed_branch() const
{
  assert (!cur_status_.empty());

size_t id;
const bool rc = get_next_id_to_be_processed (id);
  assert (rc);
  assert (id < cur_status_.size());

  return cur_status_[id].the_branch_;
}



// =========
string NondeterministicPostMachine::getstr_cur_branch() const
{
  assert (!cur_status_.empty());
vector<int> branch (get_cur_branch());
ostringstream oss;
  copy (branch.begin(), branch.end(), ostream_iterator<int> (oss, "-"));

const string str (oss.str());
  return str.substr (0, str.size() - 1); 
}



// =========
string NondeterministicPostMachine::getstr_processed_branch() const
{
  assert (!cur_status_.empty());
vector<int> branch (get_processed_branch());
ostringstream oss;
  copy (branch.begin(), branch.end(), ostream_iterator<int> (oss, "-"));

const string str (oss.str());
  return str.substr (0, str.size() - 1); 
}




// =========
bool NondeterministicPostMachine::get_next_id_to_be_processed (size_t& id_o) const
{
  assert (!cur_status_.empty());
  id_o = MachineStatus::next_id_to_be_processed_s;
  return (MachineStatus::next_id_to_be_processed_s < cur_status_.size());
}


// =========
// =========
bool NondeterministicPostMachine::perform_step (
		const command_t& command_i, 
		const vector<symbol_t>& scanned_symbols_i, 
		Commands_t::const_iterator& find_iter_o,
		const string& msg_i
		)
{
  IF_NOT_EMPTY (msg_i, 5, '=');
  assert (command_i > 0);

find_iter_o = commands_.find (command_i);
  assert (find_iter_o != commands_.end());	
  assert (command_i > 0);
  assert (command_i <= commands_.size());
  
size_t dist = distance (commands_.begin(), find_iter_o); // XXX
  assert (dist < commands_.size());
  assert (dist < command_statistics_.size());
  command_statistics_[dist]++;

  // -----------------------
  // -----------------------
const Tapes_t last_tapes (get_processed_tapes());
const vector<int> last_branch (get_processed_branch());

const size_t next_configuration_no = get_cur_configuration () + 1;

  for (size_t id = 0; id < find_iter_o->second.size(); id++)
  {
    CommandType       cur_command_type;
    vector<symbol_t>  new_symbols;
    vector<shift_t>   new_shifts;
    command_t         next_command_no;

    if (!execute_command (
			scanned_symbols_i,
			find_iter_o->second[id],
			cur_command_type,
			new_symbols,
			new_shifts,
			next_command_no
			)
			)
    {
      return false;
    }

    // -------------------------
    const int subno = id + 1;
    cur_status_.push_back (MachineStatus());

    cur_status_.back().the_branch_ = last_branch;
    cur_status_.back().the_branch_.push_back(subno);
    time_used_ = MAX_VALUE (time_used_, cur_status_.back().the_branch_.size());

    get_cur_configuration() = next_configuration_no;

    Tapes_t::iterator pos_iter;
    for (pos_iter = get_cur_tapes().begin(); pos_iter != get_cur_tapes().end(); pos_iter++)
    {
      pos_iter->second.clear_it();
    }
    get_cur_tapes().clear();

    get_cur_tapes() = last_tapes;
    // -----------------------

    get_cur_command_iter() = find_iter_o;
    get_cur_command_subno() = subno;
    get_cur_command_no() = next_command_no;

    assert (new_symbols.size() == get_cur_tapes().size());
    assert (new_shifts.size() == get_cur_tapes().size());
    for (size_t i = 0; i < get_cur_tapes().size(); i++)
    {
      get_cur_tapes()[i].set_symbol (new_symbols[i]);
      get_cur_tapes()[i].shift_position (new_shifts[i]);
    }
 
    assert (get_cur_command_no() > 0);
    assert (get_cur_command_no() <= commands_.size());

  }

  MachineStatus::next_id_to_be_processed_s++;

  return true;

}



// =========
bool NondeterministicPostMachine::analyze_simple_seq_command (
			const symbol_t&		scanned_symbol_i,
			const OperationType&    operation_i
			)
{
bool ret_bool = true;

  switch (operation_i)
  {
    case STAY_ON_op  :
    case LEFT_SHIFT_op  :
    case RIGHT_SHIFT_op :
      // Do nothing
      break;

    case WRITE_LABEL_op  :
      if (scanned_symbol_i == symbol_t (LABEL_ALPHABET_SYMBOL)) ret_bool = false;
      break;

    case ERASE_LABEL_op  :
      if (scanned_symbol_i == symbol_t (EMPTY_ALPHABET_SYMBOL)) ret_bool = false;
      break;

    default :
      assert (0);
      break; // unused
  }

  return ret_bool;

}


// =========
bool NondeterministicPostMachine::analyze_complex_seq_command (
			const vector<symbol_t>&       scanned_symbols_i,
			const vector<OperationType>&  operations_i,
			vector<symbol_t>&             new_symbols_o,
			vector<shift_t>&              new_shifts_o
			)
{
const size_t total_tapes = get_cur_tapes().size();

  assert (scanned_symbols_i.size() == total_tapes);
  assert (operations_i.size()      == total_tapes);


  new_symbols_o.clear();
  new_shifts_o.clear();


  for (size_t k = 0; k < total_tapes; k++)
  {
    const symbol_t      the_symbol    = scanned_symbols_i[k];
    const OperationType the_operation = operations_i[k];
    if (!analyze_simple_seq_command (the_symbol, the_operation))
    {
      return false;
    }


    switch (the_operation)
    {
      case STAY_ON_op  :
        new_symbols_o.push_back(the_symbol);
        new_shifts_o.push_back(NO_SHIFT);
        break;

      case LEFT_SHIFT_op  :
        new_symbols_o.push_back(the_symbol);
        new_shifts_o.push_back(LEFT_SHIFT);
        break;

      case RIGHT_SHIFT_op :
        new_symbols_o.push_back(the_symbol);
        new_shifts_o.push_back(RIGHT_SHIFT);
        break;
  
      case WRITE_LABEL_op  :
        assert (the_symbol == symbol_t (EMPTY_ALPHABET_SYMBOL));
        new_symbols_o.push_back(LABEL_ALPHABET_SYMBOL);
        new_shifts_o.push_back(NO_SHIFT);
        break;
  
      case ERASE_LABEL_op  :
        assert (the_symbol == symbol_t (LABEL_ALPHABET_SYMBOL));
        new_symbols_o.push_back(EMPTY_ALPHABET_SYMBOL);
        new_shifts_o.push_back(NO_SHIFT);
        break;
  
      default :
        assert (0);
        break; // unused
    }

  } // for (k = 
  

  assert (new_symbols_o.size() == total_tapes);
  assert (new_shifts_o.size()  == total_tapes);
  return true;
}



// =========
bool NondeterministicPostMachine::analyze_simple_cond_command (
			const symbol_t&		scanned_symbol_i,
			const OperationType&    operation_i
			)
{
  switch (operation_i)
  {
    case CHECK_IF_MARKED_op :
      if (scanned_symbol_i == symbol_t (LABEL_ALPHABET_SYMBOL)) return true;
      if (scanned_symbol_i == symbol_t (EMPTY_ALPHABET_SYMBOL)) return false;
      break;

    case CHECK_IF_BLANK_op  :
      if (scanned_symbol_i == symbol_t (LABEL_ALPHABET_SYMBOL)) return false;
      if (scanned_symbol_i == symbol_t (EMPTY_ALPHABET_SYMBOL)) return true;
      break;

    default :
      assert (0);
      break; // unused
  }

  assert (0);
  return false;	 // unused

}


// =========
bool NondeterministicPostMachine::analyze_complex_cond_command (
			const vector<symbol_t>&       scanned_symbols_i,
			const vector<OperationType>&  operations_i,
			vector<symbol_t>&             new_symbols_o,
			vector<shift_t>&              new_shifts_o
			)
{

bool ret_bool = true;
const size_t total_tapes = get_cur_tapes().size();

  assert (scanned_symbols_i.size() == total_tapes);
  assert (operations_i.size()      == total_tapes);


  new_symbols_o.clear();
  new_shifts_o.clear();


  for (size_t k = 0; k < total_tapes; k++)
  {
    const symbol_t      the_symbol    = scanned_symbols_i[k];
    const OperationType the_operation = operations_i[k];
    switch (the_operation)
    {
      case CHECK_IF_MARKED_op :
      case CHECK_IF_BLANK_op  :
        if (!analyze_simple_cond_command (the_symbol, the_operation)) ret_bool = false;
        break;
  
      default :
        assert (0);
        break; // unused
    }

  } // for (k = 
  

  copy (scanned_symbols_i.begin(), scanned_symbols_i.end(), back_inserter (new_symbols_o));
  for (size_t k = 0; k < total_tapes; k++)
  {
    new_shifts_o.push_back(NO_SHIFT);
  }

  assert (new_symbols_o.size() == total_tapes);
  assert (new_shifts_o.size()  == total_tapes);

  return ret_bool;
}


// =========
bool NondeterministicPostMachine::execute_command (
			const vector<symbol_t>& scanned_symbols_i,
			const CommandAction&    command_action_i,
			CommandType&		cur_command_type_o,
			vector<symbol_t>&       new_symbols_o,
			vector<shift_t>&        new_shifts_o,
			command_t&		next_command_no_o
			)
{
bool ret_bool = true;

  assert (scanned_symbols_i.size() == command_action_i.cur_operations_.size());

  new_symbols_o.clear();
  new_shifts_o.clear();

  assert (scanned_symbols_i.size() == command_action_i.cur_operations_.size());
  cur_command_type_o = command_action_i.get_command_type();

  switch (cur_command_type_o)
  {
    case SEQUENTIAL_cmd :
      assert (command_action_i.next_commands_.size() == 1);
      ret_bool = analyze_complex_seq_command (
			scanned_symbols_i,
			command_action_i.cur_operations_,
			new_symbols_o,
			new_shifts_o
			);
      next_command_no_o = command_action_i.next_commands_.front();
      break;

    case CONDITIONAL_JUMP_cmd :
      assert (command_action_i.next_commands_.size() == 2);
      if (analyze_complex_cond_command (
			scanned_symbols_i,
			command_action_i.cur_operations_,
			new_symbols_o,
			new_shifts_o
			)
			)
      {
        next_command_no_o = command_action_i.next_commands_.front();
      }
      else
      {
        next_command_no_o = command_action_i.next_commands_.back();
      }
 
      break;

    case END_cmd :
      assert (command_action_i.next_commands_.empty());
      break;

    default :
      assert (0);
      break; // unused
  }

  return ret_bool;

}


// =========
bool NondeterministicPostMachine::process_input (
		const vector<vector<symbol_t> >& input_words_i,
		const string& msg_i
		)
{
  IF_NOT_EMPTY (msg_i, 5, '=');

  if (!check_results_)
  {
    FATAL_MSG	("Invalid Nondeterministic Post Machine definition");
    return false;
  }

vector<vector<symbol_t> > input_words (input_words_i);
vector<long> init_pos;

  cout << endl;
  if (!check_input_words(input_words, init_pos))
  {
    show_input (input_words, init_pos, false);
    FATAL_MSG	("Invalid input words");
    cur_input_set_no_++;
    return false;
  }
  clear_it ();
  assert (input_words.size() == init_pos.size());
  show_input (input_words, init_pos);
  set_input (input_words, init_pos);


  // ------------------
command_t          command_no;
vector<symbol_t>   scanned_symbols;

bool	ret_bool_value;

  cout << endl;
  cout << endl;
  cout << "\t"
       << string (5, '#') 
       << " "
       << getstr_id_info ()
       << "Processing"
       << " "
       << string (5, '#') 
       << endl;

Commands_t::const_iterator find_iter = commands_.end();

size_t next_id;

  for (int step_no = 0; ; step_no++)
  {

    const int actual_step_no = step_no;
    if (!get_next_id_to_be_processed (next_id)) break;
    assert (next_id < cur_status_.size());

    scanned_symbols.clear();
    for (size_t i = 0; i < cur_status_[next_id].the_tapes_.size(); i++)
    {
      scanned_symbols.push_back(cur_status_[next_id].the_tapes_[i].get_scanned_symbol ());
    }

    // -------------------------------------
    show2_situation (getstr_id_info (), actual_step_no);
    ret_bool_value = perform_step (cur_status_[next_id].the_command_no_, scanned_symbols, find_iter);  

    command_no = get_processed_command_no();

    // -------------------
    show1_situation (find_iter);

    // XXX assert (((find_iter != commands_.end()) && (ret_bool_value)) || ((find_iter == commands_.end()) && (!ret_bool_value)));

    if (!ret_bool_value)
    { 
      cout << "\t" << getstr_id_info () << "Failure" << (is_actually_nondeterministic_ ? " of the Branch" : "") << endl;
      cout << "\t" << "  1) Unable to apply this command (" << command_no << ") for" << endl;
      cout << "\t" << "     [ ";
      copy (scanned_symbols.begin(), scanned_symbols.end(), ostream_iterator<string> (cout, " "));
      cout << "]" << endl;
      // cout << "\t" << "  2) Current command (" << command_no << ") is not halting one" << endl;
      cout << "\t" << "  2) Halting command is not reached" << endl;

      cout << endl;
      cout << endl;

      MachineStatus::next_id_to_be_processed_s++;
      continue;
    }

    assert (ret_bool_value);
    // -------------------
    // show2_situation (getstr_id_info (), "Configuration#" + to_string (actual_step_no));

    // -------------------
    if (is_halting_command (command_no))
    {

      show2_situation (getstr_id_info (), actual_step_no + 1);

      find_iter = commands_.find (command_no);
      assert (find_iter != commands_.end());

      show1_situation (find_iter);

      cout << "\t" << getstr_id_info () << "Success : Current command (" << command_no << ") is halting one" << endl;
      cout << endl;
      cout << endl;

      MachineStatus::next_id_to_be_processed_s++;
      break;
    }

  } // for (step_no = 

const size_t the2_setw = 3;

ostringstream oss;
  oss << "" 
       << string (the2_setw, '-')
       << " "
       << getstr_id_info (true)
       << " Result : Input word(s) " 
       << (ret_bool_value ? "ACCEPTED" : "REJECTED") 
       << " " 
       << string (the2_setw, '-'); 
  
const size_t the1_setw = oss.str().size();

const string prefix (3, ' ');

  cout << endl;
  cout << prefix << string (the1_setw, '-') << endl;
  cout << prefix 
       << oss.str()
       << endl;
  cout << prefix << string (the1_setw, '-') << endl;
  cout <<  endl;

  // -----------------------------
  assert (output_size_ == 0);
  output_size_ = get_output_size();

  assert (space_used_ == 0);
  space_used_ = get_space_used();
  
  show_resources_report ();

  // -----------------------------
  show_statistics ();

  return ret_bool_value;
}


// =========
void NondeterministicPostMachine::show_env () const
{
  cout << endl;
  cout << "\t###### Nondeterministic Post Machine Definition ######" << endl;
  if (is_actually_nondeterministic_)
  {
    cout << "\t###### This Machine is actually Nondeterministic #####" << endl;
  }
  else
  {
    cout << "\t###### This Machine is actually Deterministic!!! #####" << endl;
  }

  cout << endl;
  show_descr ();

  cout << endl;
  cout << endl;
  show_commands ();

  cout << endl;
  cout << endl;
  show_alphabets();

  cout << endl;
  cout << endl;
  show_command ();

}


// =========
void NondeterministicPostMachine::show_descr (const string& msg_i) const
{
  cout << "\t   ====== Description ======" << endl;

  IF_NOT_EMPTY (msg_i, 5, '=');

  if (descr_.empty())
  {
    cout << "No description" << endl;
  }

  for (size_t i = 0; i < descr_.size(); i++)
  {
    copy (descr_[i].begin(), descr_[i].end(), ostream_iterator<string> (cout, " "));
    cout << endl;
  }

}


// =========
void NondeterministicPostMachine::show_commands (const string& msg_i) const
{
  cout << "\t   ====== Command Numeration ======" << endl;

  IF_NOT_EMPTY (msg_i, 5, '=');


string text_initial_commands ("Initial commands");
string text_halting_commands ("Halting commands");
string text_internal_commands ("Internal commands");
size_t text_max_size = 0;

  text_max_size	= MAX_VALUE(text_max_size, text_initial_commands.size());
  text_max_size	= MAX_VALUE(text_max_size, text_halting_commands.size());
  text_max_size	= MAX_VALUE(text_max_size, text_internal_commands.size());


  cout << setw(text_max_size) << left << text_initial_commands.c_str() << " : ";
  for (size_t i = 0; i < initial_commands_.size(); i++)
  {
    cout << setw (max_command_size_) << initial_commands_[i] << " ";
  }
  cout << endl;

  cout << setw(text_max_size) << left << text_halting_commands.c_str() << " : ";
  for (size_t i = 0; i < halting_commands_.size(); i++)
  {
    cout << setw (max_command_size_) << halting_commands_[i] << " ";
  }
  cout << endl;

  cout << setw(text_max_size) << left << text_internal_commands.c_str() << " : ";
  for (size_t i = 0; i < internal_commands_.size(); i++)
  {
    cout << setw (max_command_size_) << internal_commands_[i] << " ";
  }
  cout << endl;


}

// =========
void NondeterministicPostMachine::show_alphabet (const Tapes_t::const_iterator& iter_i) const
{
  iter_i->second.show_alphabet();
}

// =========
void NondeterministicPostMachine::show_alphabets () const
{
Tapes_t::const_iterator pos_iter;
  cout << "\t   ====== Alphabet Definition ======" << endl;
  for (pos_iter = init_tapes_.begin(); pos_iter != init_tapes_.end(); pos_iter++)
  {
    cout << "\t      ------ Tape# " << distance (init_tapes_.begin(), pos_iter) << " ------" << endl;
    show_alphabet (pos_iter);
    cout << endl;
  }
}

// =========
void NondeterministicPostMachine::show_command () const
{
Commands_t::const_iterator pos_iter;
  cout << "\t   ====== Command Definition ======" << endl;
  for (pos_iter = commands_.begin(); pos_iter != commands_.end(); pos_iter++)
  {
    cout << getstr_rule_S (
		is_actually_nondeterministic_, 
		commands_.begin(), 
		pos_iter, 
		max_command_size_, 
		CommandAction::get_op_names_max_size_S ()
		);
  }
  cout << endl;
}

// =========
string NondeterministicPostMachine::getstr_rule_S (
		bool is_actually_nondeterministic_i,
		Commands_t::const_iterator iter1_i,
		Commands_t::const_iterator iter2_i,
		size_t	max_command_size_i,
		size_t	max_operation_name_size_i,
		int	subno_i
		)
{
  assert (subno_i != 0);
  assert ((subno_i > 0) || (subno_i == -1));

ostringstream oss;
  for (size_t k = 0; k < iter2_i->second.size(); k++)
  {
    if ((subno_i == -1)	|| (size_t(subno_i) == (k + 1)))
    {
      oss << ""
          << "Command#"
          << setw (SETW_RULE_NUMBER)
          << right
          << (distance (iter1_i, iter2_i) + 1);
      if (is_actually_nondeterministic_i)
      {
        oss << "-"
            << (k + 1);
      }
      oss << " :   "
          << setw (max_command_size_i)
          << right
          << iter2_i->first
          << " ---> "
          << iter2_i->second[k].getstr_cur_operation_names(max_operation_name_size_i)
          << "" 
          << iter2_i->second[k].getstr_next_commands(max_command_size_i)
          << endl;
    }
  }

  return oss.str();
}
   
// =========
void NondeterministicPostMachine::show_tape (const Tapes_t::const_iterator& iter_i) const
{
  iter_i->second.show_tape();
}


// =========
void NondeterministicPostMachine::show1_situation (
	const Commands_t::const_iterator& find_iter_i,
	bool init_configuration_i
	) const
{
  assert (!cur_status_.empty());
  assert (!init_configuration_i);

  if (find_iter_i != commands_.end())
  {
    for (size_t k = 0; k < find_iter_i->second.size(); k++)
    {
      cout << "\t" 
           << getstr_id_info ()
           << "Applied " 
           << getstr_rule_S (
		is_actually_nondeterministic_, 
		commands_.begin(), 
		find_iter_i, 
		max_command_size_, 
		CommandAction::get_op_names_max_size_S (), 
		k + 1
		); 
    }
  }
  else
  {
      cout << "\t" 
           << getstr_id_info ()
           << "No Applied Command"
           << endl;
  }
  cout << endl;


}



// =========
void NondeterministicPostMachine::show2_situation (
	const string& msg_id_info_i, 
	size_t        configuration_no_i
	) const
{
Tapes_t::const_iterator pos_iter;
const size_t size1 = 5;

  assert (!cur_status_.empty());

  if (configuration_no_i)
  {
    cout << endl;
    if (is_actually_nondeterministic_)
    {
      cout << "\t" << msg_id_info_i << "Branch : " << getstr_processed_branch() << endl;
    }
  }
const string configuration_info (configuration_no_i ? ("Configuration#" + to_string(configuration_no_i)) : "Initial Configuration");
  IF_NOT_EMPTY ((msg_id_info_i + configuration_info), size1, '-');


  // cout << "\tState  : " << get_processed_command_no() << endl;
  //cout << "\t" << string (size2, '-') << endl;

const Tapes_t the_tapes (get_processed_tapes());
  for (pos_iter = the_tapes.begin(); 
       pos_iter != the_tapes.end(); 
       pos_iter++
      )
  {
    cout << "Tape#" << distance (the_tapes.begin(), pos_iter) << " : ";
    show_tape (pos_iter);
  }
  //cout << endl;

}


// =========
void NondeterministicPostMachine::show_resources_report (const string& msg_i) const
{
const size_t size1 = 5;

  IF_NOT_EMPTY (msg_i, size1, '-');

ostringstream oss, oss2;
const string prefix (3, ' ');
const string li_sign (" * ");
const size_t info_setw = 10;

  oss2 << getstr_id_info (true) << " Resource Report";

  oss << "" 
      << prefix
      << oss2.str()
      << endl

      << prefix
      << string (oss2.str().size(), '-')
      << endl

      << prefix
      << getstr_id_info (true)
      << li_sign
      << "Input size    : "
      << setw (info_setw)
      << std::right
      << input_size_
      << " symbols"
      << endl

      << prefix
      << getstr_id_info (true)
      << li_sign
      << "Output size   : "
      << setw (info_setw)
      << std::right
      << output_size_
      << " symbols"
      << endl

      << prefix
      << getstr_id_info (true)
      << li_sign
      << "PM-Space used : "
      << setw (info_setw)
      << std::right
      << space_used_
      << " cells"
      << endl

      << prefix
      << getstr_id_info (true)
      << li_sign
      << "PM-Time used  : "
      << setw (info_setw)
      << std::right
      << time_used_;
  if (is_actually_nondeterministic_)
  {
    oss << " (smart)";
  }
  oss << " commands"
      << endl;


  cout << oss.str() << endl << endl << endl;

}


// =========
void NondeterministicPostMachine::show_statistics (const string& msg_i) const
{
const size_t size1 = 5;

  IF_NOT_EMPTY (msg_i, size1, '-');

uint total = accumulate (command_statistics_.begin(), command_statistics_.end(), 0);

const string text_statistics ("Statistics");
const string text_command    (" Commands ");
const string text_rules ("Command");
const string text_times ("Times");
const string text_total ("Total");
const string delim (":");

const size_t max_size1 = MAX_VALUE (text_statistics.size(), text_command.size());
const size_t the_size2 = 4;
const size_t the_size1 = the_size2 + 1 + max_size1 + 1 + the_size2;
const size_t the_size3 = (the_size1 - 1)/2;
const size_t the_size4 = (the_size1 - 1) - the_size3;
  assert (text_rules.size() < (the_size3 - 1));
  assert (text_times.size() < (the_size4 - 1));

const string l_border ("\t      |");		
const string r_border ("|");		

  cout << l_border << string (the_size1, '=') << r_border << endl;

  cout << l_border 
       << string (the_size2, '-') 
       << " "
       << setw (max_size1)
       << left
       << text_statistics.c_str()
       << " "
       << string (the_size2, '-') 
       << r_border 
       << endl;

  cout << l_border 
       << string (the_size2, '.') 
       << " "
       << setw (max_size1)
       << left
       << text_command.c_str()
       << " "
       << string (the_size2, '.') 
       << r_border 
       << endl;

  cout << l_border << string (the_size1, '-') << r_border << endl;

  cout << l_border 
       << setw (the_size3 - 1)
       << right
       << text_rules.c_str()
       << " "
       << delim
       << setw (the_size4 - 1)
       << right
       << text_times.c_str()
       << " "
       << r_border 
       << endl;

  cout << l_border << string (the_size1, '-') << r_border << endl;

  for (size_t i = 0; i < command_statistics_.size(); i++)
  {
    cout << l_border 
         << setw (the_size3 - 1)
         << right
         << i
         << " "
         << delim
         << setw (the_size4 - 1)
         << right
         << command_statistics_[i]
         << " "
         << r_border 
         << endl;
  }
  cout << l_border << string (the_size1, '-') << r_border << endl;

  cout << l_border 
       << setw (the_size3 - 1)
       << right
       << text_total.c_str()
       << " "
       << delim
       << setw (the_size4 - 1)
       << right
       << total
       << " "
       << r_border 
       << endl;

  cout << l_border << string (the_size1, '=') << r_border << endl;

}


// =========
bool NondeterministicPostMachine::is_initial_command (const command_t& command_i) const
{
  return (find (initial_commands_.begin(), initial_commands_.end(), command_i) != initial_commands_.end());
}

// =========
bool NondeterministicPostMachine::is_halting_command (const command_t& command_i) const
{
  return (find (halting_commands_.begin(), halting_commands_.end(), command_i) != halting_commands_.end());
}


// =========
bool NondeterministicPostMachine::is_internal_command (const command_t& command_i) const
{
  return (find (internal_commands_.begin(), internal_commands_.end(), command_i) != internal_commands_.end());
}

// =========
bool NondeterministicPostMachine::is_valid_command (const command_t& command_i) const
{
  assert ((size_t(command_i) >= 1) && (size_t(command_i) <= commands_.size()));
  return (is_initial_command(command_i) || is_halting_command (command_i) || is_internal_command (command_i));
}

// =========
bool NondeterministicPostMachine::is_valid_operation (const OperationType& operation_no_i) const
{
  return ((operation_no_i < LAST_OperationType) && (operation_no_i != UNDEF_OperationType));
}



// =========
bool NondeterministicPostMachine::check_commands () const
{
bool	ret_bool_value = true;

  // ---------
  if (initial_commands_.empty())
  {
    ret_bool_value = false;
    FATAL_MSG	("Missing initial command"
		);
  }

  if (halting_commands_.empty())
  {
    ret_bool_value = false;
    FATAL_MSG	("Set of halting command is empty"
		);
  }

  // -----------
  return ret_bool_value;
} // check_commands



// =========
bool NondeterministicPostMachine::check_command () const
{
bool	ret_bool_value = true;

Commands_t::const_iterator	pos1_iter;
Tapes_t::const_iterator		pos2_iter;
command_t	cur_command;
symbol_t	cur_symbol;

symbol_t	next_symbol;
shift_t		shift;

  if (commands_.empty())
  {
    ret_bool_value = false;
    FATAL_MSG	("No command function");
  }

  for (pos1_iter = commands_.begin(); pos1_iter != commands_.end(); pos1_iter++)
  {
    const string commands_line_info ("Command Line#" + to_string (distance (commands_.begin(), pos1_iter)));
    // const string commands_line_and_tape_no_info (commands_line_info + ", tape#");

    // --- first ---
    cur_command = pos1_iter->first;

    if (!is_valid_command(cur_command))
    {
      ret_bool_value = false;
      FATAL_MSG	(commands_line_info
		<< " : illegal cur-command = <"
		<< cur_command
		<< ">"
		);
    }

    
    // --- second ---
    for (size_t k = 0; k < pos1_iter->second.size(); k++)
    {

      const string commands_line_second_info (commands_line_info + "-" + to_string (k + 1));
      // const string commands_line_and_tape_no_second_info (commands_line_second_info + ", tape#");

      for (size_t s = 0; s < pos1_iter->second[k].cur_operations_.size(); s++)
      {
        const OperationType the_operation = pos1_iter->second[k].cur_operations_[s];

        if (!is_valid_operation(the_operation))
        {
          ret_bool_value = false;
          FATAL_MSG	(commands_line_second_info
			<< " : illegal operation = <"
			<< the_operation
			<< ">"
			);
        }

      } // for (size_t s = 0;

      for (size_t s = 0; s < pos1_iter->second[k].next_commands_.size(); s++)
      {
        const command_t the_next_command = pos1_iter->second[k].next_commands_[s];

        if (!is_valid_command(the_next_command))
        {
          ret_bool_value = false;
          FATAL_MSG	(commands_line_second_info
			<< " : illegal next-command = <"
			<< the_next_command
			<< ">"
			);
        }

      } // for (size_t s = 0;

    } //for (size_t k = 0
 
  } // for (iter = command.begin(); ...
  return ret_bool_value;
}

	
// =========
void NondeterministicPostMachine::detect_if_is_actually_nondterministic ()
{
Commands_t::const_iterator	pos_iter;
  for (pos_iter = commands_.begin(); pos_iter != commands_.end(); pos_iter++)
  {
    assert (pos_iter->second.size() > 0);
    if (pos_iter->second.size() > 1)
    {
      is_actually_nondeterministic_ = true;
      break;
    }
  }

}


// =========
bool NondeterministicPostMachine::get_check_results () const
{
  return check_results_;
}

// =========
string NondeterministicPostMachine::getstr_id_info (bool detail_flag_i) const
{
ostringstream oss;
  oss << "< ";
  if (detail_flag_i)
  {
    oss << (is_actually_nondeterministic_ ? "N" : "D")
        << "PM #"
        << serial_no_ 
        << ", "
        << "Input #"
        << cur_input_set_no_;
  }
  else
  {
    oss << "Run "
        << serial_no_ 
        << "."
        << cur_input_set_no_;
  }
  oss << " > ";

  return oss.str();
}


