// ==============================================================
//
//  Copyright (c) 2002-2003 Alex Vinokur.
//
//  ------------------------------------------------------------
//  This file is part of C++ Simulator of a Post Machine.
//
//  C++ Simulator of a Post Machine is free software; 
//  you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License,
//  or (at your option) any later version.
//
//  C++ Simulator of a Post Machine is distributed in the hope 
//  that it will be useful, but WITHOUT ANY WARRANTY; 
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with C++ Simulator of a Post Machine; 
//  if not, write to the Free Software Foundation, Inc., 
//  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ------------------------------------------------------------
// 
//  mailto:alexvn@connect.to
//  http://up.to/alexv
//
// ==============================================================




// ##############################################################
//
//  SOFTWARE : Nondeterministic Post Machine (C++ Simulator)
//  FILE     : post-m.h
//
//  DESCRIPTION :
//         Class NondeterministicPostMachine (Definition)
//
// ##############################################################




// ###############
#ifndef _POST_M_H
#define _POST_M_H


#include "commands.h"
#include "tape.h"


typedef map<command_t, vector<CommandAction> >		Commands_t;
typedef map<size_t, Tape>				Tapes_t;


// =========
class MachineStatus
{
friend class NondeterministicPostMachine;
  private :
    vector<int>				the_branch_;
    size_t				the_configuration_no_;
    Commands_t::const_iterator		the_command_iter_;
    int					the_command_subno_;
    command_t				the_command_no_;
    Tapes_t				the_tapes_;

    static size_t			next_id_to_be_processed_s;

  public :
    MachineStatus ();
    ~MachineStatus ();
    

};

// =========
class NondeterministicPostMachine
{
  private :
    // ------ Private Data ------
    bool			is_actually_nondeterministic_;
    const vector<vector<string>	> descr_;
    const vector<command_t>	initial_commands_;
    const vector<command_t>	halting_commands_;
    const vector<command_t>	internal_commands_;
    const Commands_t		commands_;
    bool			check_results_;

    vector<MachineStatus>	cur_status_;
    size_t			cur_input_set_no_;
    Tapes_t			init_tapes_;
    vector<uint>  		command_statistics_;

    size_t	max_command_size_;
    size_t	max_symbol_size_;

    static size_t	counter_s;
    const size_t	serial_no_;

    size_t	input_size_;
    ulong	output_size_;
    ulong	space_used_;
    ulong	time_used_;

    // ------ Private Methods ------
    void	clear_it ();
    void	set_input (
			const vector<vector<symbol_t> >& input_words_i,
			const vector<long>&              init_pos_i
			);
    bool	perform_step (
			const command_t& command_i, 
			const vector<symbol_t>& scanned_symbols_i, 
			Commands_t::const_iterator& find_iter_o,
			const string& msg_i = string()
			);

    bool	execute_command (
			const vector<symbol_t>& scanned_symbols_i,
			const CommandAction&    command_action_i,
			CommandType&            cur_command_type_o,
			vector<symbol_t>&       new_symbols_o,
			vector<shift_t>&        new_shifts_o,
			command_t&              next_command_no_o
			);

    bool	analyze_simple_seq_command (
			const symbol_t&       scanned_symbol_i,
			const OperationType&  operation_i
			);


    bool	analyze_complex_seq_command (
			const vector<symbol_t>&       scanned_symbols_i,
			const vector<OperationType>&  operations_i,
			vector<symbol_t>&             new_symbols_o,
			vector<shift_t>&              new_shifts_o
			);


    bool	analyze_simple_cond_command (
			const symbol_t&		scanned_symbol_i,
			const OperationType&    operation_i
			);


    bool	analyze_complex_cond_command (
			const vector<symbol_t>&       scanned_symbols_i,
			const vector<OperationType>&  operations_i,
			vector<symbol_t>&             new_symbols_o,
			vector<shift_t>&              new_shifts_o
			);


    vector<command_t>	get_all_commands () const;
    string		getstr_id_info (bool detail_flag_i = false) const;

    size_t&		get_cur_configuration ();
    size_t		get_processed_configuration () const;

    Tapes_t&		get_cur_tapes ();
    Tapes_t		get_processed_tapes () const;

    command_t&		get_cur_command_no ();
    command_t		get_processed_command_no () const;


    Commands_t::const_iterator&	get_cur_command_iter();
    Commands_t::const_iterator	get_processed_command_iter() const;

    int&		get_cur_command_subno();
    int			get_processed_command_subno() const;

    vector<int>		get_cur_branch () const;
    string		getstr_cur_branch () const;

    vector<int>		get_processed_branch () const;
    string		getstr_processed_branch () const;


    bool		get_next_id_to_be_processed  (size_t& id_o) const;

    ulong		get_output_size ();
    ulong		get_space_used ();

    bool	is_initial_command (const command_t& command_i) const;
    bool	is_halting_command (const command_t& command_i) const;
    bool	is_internal_command (const command_t& command_i) const;
    bool	is_valid_command (const command_t& command_i) const;
    bool	is_valid_operation (const OperationType& operation_no_i) const;

    void 	set_max_command_size();
    void	set_max_symbol_size();

    bool	check_commands () const;
    bool	check_command () const;
    void	detect_if_is_actually_nondterministic ();
    bool	check_input_words (
			vector<vector<symbol_t> >& input_words_io, 
			vector<long>& init_pos_o
			) const;

    void	show_descr (const string& msg_i = string()) const;
    void	show_commands (const string& msg_i = string()) const;
    void	show_alphabet (const Tapes_t::const_iterator& iter) const;
    void	show_alphabets () const;
    void	show_command () const;
    static string	getstr_rule_S (
			bool is_actually_nondeterministic_i,
			Commands_t::const_iterator iter1_i,
			Commands_t::const_iterator iter2_i,
			size_t max_command_size_i,
			size_t max_operation_name_size_i,
			int    subno_i = -1
			);
    void	show_tape (const Tapes_t::const_iterator& iter) const;
    void	show1_situation (
			const Commands_t::const_iterator& find_iter_i,
			bool init_configuration_i = false
			) const;
    void	show2_situation (
			const string& msg_id_info_i, 
			size_t        configuration_no_i
			) const;

    void	show_statistics (const string& msg_i = string()) const;
    void	show_resources_report (const string& msg_i = string()) const;

    void	show_input (
			const vector<vector<symbol_t> >& input_words_i, 
			const vector<long>&              init_pos_i,
			bool                             full_flag_i = true
			) const;


  public :
    NondeterministicPostMachine (
	size_t                         number_of_tapes_i,
	const vector<vector<string> >& descr_i, 
	const vector<command_t>&       initial_command_numbers_i,
	const vector<command_t>&       halting_command_numbers_i,
	const vector<command_t>&       internal_command_numbers_i,
	const Commands_t&              commands_i,
	const string&                  msg_i = string ()
	);
    ~NondeterministicPostMachine ();  
    bool	process_input (
			const vector<vector<symbol_t> >& input_words_i,
			const string& msg_i = string()
			);
    void	show_env () const;
    bool	get_check_results () const;

};

#endif	// _POST_M_H
